/*
 * Copyright (c) 2012-2014 Wind River Systems, Inc.
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <stdio.h>
#include <zephyr/kernel.h>
#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/gatt.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/drivers/gpio.h>

#include <zephyr/logging/log.h>

LOG_MODULE_REGISTER(app);

#define IMU_SAMPLE_SIZE		sizeof(uint32_t)
#define IMU_SAMPLE_INT_MS	20

/* UUID of custom service */
#define BT_UUID_DUMMY_VAL BT_UUID_128_ENCODE(0x205020d0, 0xbe54, 0x48ce, 0xa4ff, 0xd992e2a85f4e)

/* UUID of notify characteristic */
#define BT_UUID_TX_VAL BT_UUID_128_ENCODE(0x205020d1, 0xbe54, 0x48ce, 0xa4ff, 0xd992e2a85f4e)

#define BT_UUID_CUSTOM_SERVICE BT_UUID_DECLARE_128(BT_UUID_DUMMY_VAL)
#define BT_UUID_CUSTOM_TX      BT_UUID_DECLARE_128(BT_UUID_TX_VAL)


struct sensor_data_t {
	void *fifo_reserved;
	uint8_t dummy_data[(CONFIG_BT_L2CAP_TX_MTU - 3)];
	uint16_t len;
};

static K_FIFO_DEFINE(fifo_sensor_data);

static struct bt_conn *current_conn;

static void tx_ccc_cfg_changed(const struct bt_gatt_attr *attr, uint16_t value);

static volatile uint32_t notification_cnt;
static volatile uint32_t notification_payload = 20;

void notification_timer_callback(struct k_timer *timer_id) 
{
	LOG_INF("%d samples per second",
		(notification_cnt * notification_payload) / IMU_SAMPLE_SIZE);
	notification_cnt = 0;
}

K_TIMER_DEFINE(notification_timer, notification_timer_callback, NULL);

static void notification_timer_start(void)
{
	k_timer_start(&notification_timer, K_SECONDS(1), K_SECONDS(1));
	LOG_INF("Timer started");
}

static void notification_timer_stop(void)
{
	k_timer_stop(&notification_timer);
	LOG_INF("Timer stopped");
}


void imu_sim_callback(struct k_timer *timer_id)
{
	struct sensor_data_t *sensor_data;

	sensor_data = k_malloc(sizeof(*sensor_data));
	if (!sensor_data) {
		LOG_WRN("Unable to allocate IMU buffer from heap");
		return;
	}
	memset(sensor_data->dummy_data, 'a', ARRAY_SIZE(sensor_data->dummy_data));
	sensor_data->len = ARRAY_SIZE(sensor_data->dummy_data);
	k_fifo_put(&fifo_sensor_data, sensor_data);
}

K_TIMER_DEFINE(imu_timer, imu_sim_callback, NULL);

static void imu_sensor_sim_start(void)
{
	k_timer_start(&imu_timer, K_MSEC(IMU_SAMPLE_INT_MS), K_MSEC(IMU_SAMPLE_INT_MS));
	LOG_INF("Timer started");
}

static void imu_sensor_sim_stop(void)
{
	k_timer_stop(&imu_timer);
	LOG_INF("Timer stopped");
}

/* Service Declaration */
BT_GATT_SERVICE_DEFINE(custom_svc, BT_GATT_PRIMARY_SERVICE(BT_UUID_CUSTOM_SERVICE),
		       BT_GATT_CHARACTERISTIC(BT_UUID_CUSTOM_TX, BT_GATT_CHRC_NOTIFY,
					      BT_GATT_PERM_READ, NULL, NULL, NULL),
		       BT_GATT_CCC(tx_ccc_cfg_changed, BT_GATT_PERM_READ | BT_GATT_PERM_WRITE), );

static const struct bt_data ad[] = {
	BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
	BT_DATA(BT_DATA_NAME_COMPLETE, CONFIG_BT_DEVICE_NAME, sizeof(CONFIG_BT_DEVICE_NAME) - 1),
};

static int bt_custom_tx_send(struct bt_conn *conn, const uint8_t *data, uint16_t len)
{
	const struct bt_gatt_attr *attr = &custom_svc.attrs[2];

	return bt_gatt_notify(conn, attr, data, len);
}

static void tx_ccc_cfg_changed(const struct bt_gatt_attr *attr, uint16_t value)
{
	if (value == BT_GATT_CCC_NOTIFY) {
		LOG_INF("Notifications enabled");
		imu_sensor_sim_start();
		notification_timer_start();
	} else {
		LOG_INF("Notifications disabled");
		imu_sensor_sim_stop();
		notification_timer_stop();
	}
}

static void connected(struct bt_conn *conn, uint8_t err)
{
	if (err) {
		LOG_INF("Connection failed (err %u)", err);
		return;
	}

	current_conn = bt_conn_ref(conn);

	LOG_INF("Connected");
}

static void disconnected(struct bt_conn *conn, uint8_t reason)
{
	LOG_INF("Disconnected (reason %u)", reason);

	if (current_conn) {
		bt_conn_unref(current_conn);
		current_conn = NULL;
	}
	imu_sensor_sim_stop();
	notification_timer_stop();
}

#if defined(CONFIG_BT_USER_PHY_UPDATE)
static void le_phy_updated(struct bt_conn *conn,
			   struct bt_conn_le_phy_info *param)
{
	LOG_INF("LE PHY Updated: Tx 0x%x, Rx 0x%x", param->tx_phy,
	       param->rx_phy);
}
#endif /* CONFIG_BT_USER_PHY_UPDATE */

#if defined(CONFIG_BT_USER_DATA_LEN_UPDATE)
static void le_data_len_updated(struct bt_conn *conn,
				struct bt_conn_le_data_len_info *info)
{
	LOG_INF("Data length updated: max tx %u (%u us) max rx %u (%u us)", info->tx_max_len,
		info->tx_max_time, info->rx_max_len, info->rx_max_time);
}
#endif /* CONFIG_BT_USER_DATA_LEN_UPDATE */

void connection_param_update(struct bt_conn *conn, uint16_t interval, uint16_t latency,
			     uint16_t timeout)
{
	float result = interval * 1.25f;
	int ci = (int)result;
	int ci_fract = (result - ci) * 100;

	LOG_INF("\r\nConnection paramaters updated.\r\n - Interval: %d.%d ms\r\n - Latency: %d\r\n - "
		"Timeout: "
		"%d ms",
		ci, ci_fract, latency, timeout * 10);
}

BT_CONN_CB_DEFINE(conn_callbacks) = {
	.connected = connected,
	.disconnected = disconnected,
	.le_param_updated = connection_param_update,
#if defined(CONFIG_BT_USER_PHY_UPDATE)
	.le_phy_updated = le_phy_updated,
#endif /* CONFIG_BT_USER_PHY_UPDATE */
#if defined(CONFIG_BT_USER_DATA_LEN_UPDATE)
	.le_data_len_updated = le_data_len_updated,
#endif /* CONFIG_BT_USER_DATA_LEN_UPDATE */
};

void att_mtu_updated(struct bt_conn *conn, uint16_t tx, uint16_t rx)
{
	notification_payload = rx - 3;
	LOG_INF("att MTU updated to %d bytes", rx);
}

static struct bt_gatt_cb gatt_cb = {
	.att_mtu_updated = att_mtu_updated,
};


int main(void)
{
	int err;

	LOG_INF("Peripheral sample with custom service starting.");

	err = bt_enable(NULL);
	if (err) {
		LOG_INF("Bluetooth init failed (err %d)", err);
		return 0;
	}

	bt_gatt_cb_register(&gatt_cb);

	err = bt_le_adv_start(BT_LE_ADV_CONN, ad, ARRAY_SIZE(ad), NULL, 0);
	if (err) {
		LOG_INF("Advertising failed to start (err %d)", err);
		return 0;
	}

	while (1) {
		/* Wait indefinitely for data to be sent over bluetooth */
		struct sensor_data_t *sensor_data = k_fifo_get(&fifo_sensor_data, K_FOREVER);

		err = bt_custom_tx_send(current_conn, sensor_data->dummy_data, sensor_data->len);
		if (err) {
			LOG_WRN("Failed to send data over BLE connection");
		} else {
			notification_cnt++;
		}

		k_free(sensor_data);
	}
}
